package core

import (
	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/ast"
	"git.akyoto.dev/cli/q/src/build/cpu"
	"git.akyoto.dev/cli/q/src/build/errors"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/token"
)

// ExpressionToRegister puts the result of an expression into the specified register.
func (f *Function) ExpressionToRegister(node *expression.Expression, register cpu.Register) error {
	if node.IsFolded {
		f.SaveRegister(register)
		f.RegisterNumber(asm.MOVE, register, node.Value)
		return nil
	}

	if node.IsLeaf() {
		return f.TokenToRegister(node.Token, register)
	}

	if ast.IsFunctionCall(node) {
		err := f.CompileCall(node)

		if register != f.CPU.Output[0] {
			f.SaveRegister(register)
			f.RegisterRegister(asm.MOVE, register, f.CPU.Output[0])
		}

		return err
	}

	if node.Token.Kind == token.Array {
		array := f.VariableByName(node.Children[0].Token.Text(f.File.Bytes))
		offset, err := f.Number(node.Children[1].Token)
		f.MemoryRegister(asm.LOAD, asm.Memory{Base: array.Register, Offset: byte(offset), Length: 1}, register)
		return err
	}

	if len(node.Children) == 1 {
		if !node.Token.IsUnaryOperator() {
			return errors.New(errors.MissingOperand, f.File, node.Token.End())
		}

		err := f.ExpressionToRegister(node.Children[0], register)

		if err != nil {
			return err
		}

		return f.ExecuteRegister(node.Token, register)
	}

	left := node.Children[0]
	right := node.Children[1]
	final := register

	if f.UsesRegister(right, register) {
		register = f.NewRegister()
	}

	err := f.ExpressionToRegister(left, register)

	if err != nil {
		return err
	}

	err = f.Execute(node.Token, register, right)

	if register != final {
		f.SaveRegister(final)
		f.RegisterRegister(asm.MOVE, final, register)
		f.FreeRegister(register)
	}

	return err
}
